#include <iostream>
#include <string>
#include <vector>
#include <map>
#include <algorithm>
#include <functional>
#include <stdexcept>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <setjmp.h>
#include <time.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <dlfcn.h>
#include <unistd.h>
#include <signal.h>

#include <bfd.h>

#ifdef __APPLE__
# define SO_SUFFIX ".dylib"
#else
# define SO_SUFFIX ".so"
#endif

using namespace std;

namespace {
    static int crash = 0;

    void die(string msg) {
        cout << msg << endl;
        exit(1);
    }

    void dump_func(char* p) {
        while (*p != (char)0xc3) {
            printf("%02x ", (*p++) & 0xff);
        }
        printf("c3\n");
    }

    template <class T_>
    void delete_fun(T_* vp) {
        delete(vp);
    }

    vector<int> get_backtrace();
    void print_backtrace(vector<int> st);

    jmp_buf jmpbuf;

    void dtr_assert_fail(const char* ass, const char* file, int line)
    {
        printf("F\n\n");
        printf("%s:%d: assertion '%s' failure!\n", file, line, ass);
        print_backtrace(get_backtrace());
        puts("");
        if (crash) abort();
        longjmp(jmpbuf, 1);
    }

    void dtr_assert_fail_old(const char* file, int line, const char* ass) {
        dtr_assert_fail(ass, file, line);
    }

    void dtr_assert_fail_d(const char* file, int line) {
        dtr_assert_fail("D assertion", file, line);
    }

    void sigtrap(int sigid, siginfo_t* si, void* uc) {
        printf("S\n\n");
        printf("signal %d occured!\n", sigid);
#if defined(__i386__) || defined(__amd64__)
        ucontext_t* uctx = (ucontext_t*)uc;
        vector<int> st = get_backtrace();
        st.erase(st.begin());
        st.erase(st.begin());
#if defined(__linux__)
#if __WORDSIZE == 64
        st.insert(st.begin(), uctx->uc_mcontext.gregs[REG_RIP]+1);
#else
        st.insert(st.begin(), uctx->uc_mcontext.gregs[REG_EIP]+1);
#endif
#elif defined(__FreeBSD__)
        st.insert(st.begin(), uctx->uc_mcontext.mc_eip+1);
#endif
        print_backtrace(st);
#else
        printf("???\n");
        print_backtrace(get_backtrace());
#endif
        puts("");
        longjmp(jmpbuf, 2);
    }

//    map<void*, string> undefined;

    void undefined_reference_fun() {
        puts("undefined reference");
        print_backtrace(get_backtrace());
        exit(1);
//        printf("%x\n", __builtin_return_address(0));
//        die(undefined[__builtin_return_address(0)] + ": undefined reference");
    }

    map<string, int> fps;
}

class DtrFile {
public:
    explicit DtrFile(string f);
    ~DtrFile();

    void load_file();
    void load_bfd();
    void load_symtab();
    void load_syms();
    void load_rels();
    void load_ctors();

    bool is_inside(void* p) {
        return (p >= image) && (p < image+image_size);
    }

    void find_nearest_line(int p, const char** file, const char** func,
                           unsigned int* line);

    int origin() {
        return (int)(image + abfd->origin);
    }

private:
    explicit DtrFile(DtrFile* a, bfd* b);

    bool is_archive() {
        return !arch.empty();
    }

    void add_sym(asymbol* asym);
    void close_bfd();

    void die_bfd(string msg) {
        bfd_perror((string(abfd->filename) + ": " + msg).c_str());
        exit(1);
    }

    void load_ctor_section(asection* sec);

private:
    string filename;
    bfd* abfd;
    char* image;
    int image_size;
    asymbol** syms;
    int symnum;

    // for archive
    DtrFile* parent;
    vector<DtrFile*> arch;

};

class DtrTest {
public:
    DtrTest(void (*fp0)(), DtrFile* file0);
    void run();

private:
    void (*fp)();
    DtrFile* file;
};

namespace {
    vector<DtrTest*> tests;
    vector<DtrFile*> files;
    int failnum;
    int errnum;
    int assertnum;
    vector<void*> dlhandles;

    vector<int> get_backtrace() {
        struct layout {
            void* fp;
            void* rp;
        };
        layout* l = (layout*)__builtin_frame_address(0);
        vector<int> st;
        while (l) {
            st.push_back((int)l->rp);
            l = (layout*)l->fp;
        }
        return st;
    }

    void print_backtrace(vector<int> st) {
        cout << "Backtrace:" << endl;

        bool bt_started = false;
        string unknown_lines;
        for (vector<int>::const_iterator ite = st.begin();
             ite != st.end(); ++ite)
        {
//            printf("%08x\n", *ite);

            vector<DtrFile*>::iterator dite =
                find_if(files.begin(), files.end(),
                        bind2nd(mem_fun(&DtrFile::is_inside), *ite));

            if (dite == files.end()) {
                if (bt_started) {
                    char buf[1024];
                    sprintf(buf, "%08x ???\n", *ite);
                    unknown_lines += string(buf);
                }
                continue;
            }

            bt_started = true;
            cout << unknown_lines;
            unknown_lines = "";

            const char* filename;
            const char* funcname;
            unsigned int lineno;
            (*dite)->find_nearest_line(*ite, &filename, &funcname, &lineno);
            if (lineno) {
                printf("%08x %s:%d: (%s)\n", *ite, filename, lineno, funcname);
            }
            else {
                printf("%08x %s:??: (%s)\n", *ite, filename, funcname);
            }
        }
    }
}

DtrFile::DtrFile(string f)
    : filename(f), abfd(0), image(0), image_size(0),
      syms(0), symnum(0), parent(0)
{}

DtrFile::DtrFile(DtrFile* a, bfd* b)
    : filename(a->filename), abfd(b),
      image(a->image), image_size(a->image_size), syms(0), symnum(0), parent(a)
{
//    printf("%s@%s %08x -\n", b->filename, filename.c_str(), image+b->origin);
}

DtrFile::~DtrFile() {
    close_bfd();
    if (!parent) free(image);
    free(syms);
    for_each(arch.begin(), arch.end(), delete_fun<DtrFile>);
}

void DtrFile::load_file() {
    FILE* fp = fopen(filename.c_str(), "rb");
    if (!fp) {
        die(filename + ": not found");
    }
    fseek(fp, 0, SEEK_END);
    image_size = ftell(fp);
    fseek(fp, 0, SEEK_SET);
    image = (char*)malloc(image_size);
    fread(image, 1, image_size, fp);
    fclose(fp);

//    printf("%s: %08x - %08x\n", filename.c_str(), image, image+image_size);
}

void DtrFile::load_bfd() {
    abfd = bfd_openr(filename.c_str(), NULL);
    if (abfd == NULL) {
        die_bfd("bfd_openr");
    }
    if (filename.find(".a") == filename.size()-2) {
        if (!bfd_check_format(abfd, bfd_archive)) {
            die_bfd("bfd_check_format");
        }

        bfd* b = 0;
        while ((b = bfd_openr_next_archived_file(abfd, b))) {
            if (!bfd_check_format(b, bfd_object)) {
                die_bfd("bfd_check_format");
            }
            arch.push_back(new DtrFile(this, b));
        }
        if (bfd_get_error() != bfd_error_no_more_archived_files) {
            die_bfd("bfd_openr_next_archived_file");
        }
        bfd_set_error(bfd_error_no_error);
    }
    else {
        if (!bfd_check_format(abfd, bfd_object)) {
            die_bfd("bfd_check_format");
        }
    }
}

void DtrFile::load_symtab() {
    if (is_archive()) {
        for_each(arch.begin(), arch.end(), mem_fun(&DtrFile::load_symtab));
        return;
    }

    if (!(bfd_get_file_flags(abfd) & HAS_SYMS)) {
        if (bfd_get_error() != bfd_error_no_error) {
            die_bfd("bfd_get_file_flags");
        }
        return;
    }

    long storage = bfd_get_symtab_upper_bound(abfd);
    if (storage < 0) {
        die_bfd("bfd_get_symtab_upper_bound");
    }
    if (storage) syms = (asymbol**)malloc(storage);

    symnum = bfd_canonicalize_symtab(abfd, syms);
    if (symnum < 0) {
        die_bfd("bfd_canonicalize_symtab");
    }
}

void DtrFile::load_syms() {
    if (is_archive()) {
        for_each(arch.begin(), arch.end(), mem_fun(&DtrFile::load_syms));
        return;
    }

    for_each(syms, syms+symnum, bind1st(mem_fun(&DtrFile::add_sym), this));
}

void DtrFile::load_rels() {
    if (is_archive()) {
        for_each(arch.begin(), arch.end(), mem_fun(&DtrFile::load_rels));
        return;
    }

//    asection* sect = bfd_get_section_by_name(abfd, ".text");
    asection* sect = abfd->sections;

    while (sect) {
        int size = bfd_get_reloc_upper_bound(abfd, sect);
        if (size < 0) {
            die_bfd("bfd_get_reloc_upper_bound");
        }

        arelent** loc = (arelent**)malloc(size);

        size = bfd_canonicalize_reloc(abfd, sect, loc, syms);
        if (size < 0) {
            die_bfd("bfd_canonicalize_reloc");
        }

        for (int i = 0; i < size; i++) {
            arelent* rel = loc[i];
/*
  char* errmsg;
  bfd_perform_relocation(abfd, rel, image, sect, NULL, &errmsg);
*/
            int* p = (int*)(origin() + sect->filepos + rel->address);
            if (!rel->sym_ptr_ptr || !*(rel->sym_ptr_ptr)) continue;
            const char* name = (*(rel->sym_ptr_ptr))->name;
            if (!name || !name[0]) continue;

            if (strncmp(name, "__assert", 8) == 0) {
                if (strlen(name) == 8) {
                    *p = (int)dtr_assert_fail_old;
                }
                else {
                    *p = (int)dtr_assert_fail;
                }
                if (rel->howto->pc_relative) *p -= (int)p+4;
                assertnum++;
            }
            /* section */
            else if (name[0] == '.') {
                asection* s = bfd_get_section_by_name(abfd, name);
                if (!s) continue; // @@@ gotpc, gotoff
                *p = origin() + *p + s->filepos + rel->addend;
            }
            /* function */
            else {
                map<string, int>::iterator ite = fps.find(name);
                if (ite != fps.end()) {
                    mprotect((void*)(((int)ite->second+4095) & ~4095 - 4096),
                             4096, PROT_READ | PROT_WRITE | PROT_EXEC);
                    *p = ite->second;
                    if (rel->howto->pc_relative) *p -= (int)p+4;
                }
                else {
                    int fpv = 0;
                    for (vector<void*>::iterator ite = dlhandles.begin();
                         ite != dlhandles.end(); ++ite)
                    {
                        fpv = (int)dlsym(*ite, name);
                        if (fpv) break;
                    }
                    if (!fpv) {
                        if (parent || name[0] == '_') {
                            fpv = (int)undefined_reference_fun;
//                            undefined[p+4] = name;
//                            printf("%s undef?\n", name);
                        }
                        else {
                            die(filename+": "+name+": undefined reference");
                        }
                    }
                    *p = fpv;
                    if (rel->howto->pc_relative) *p -= (int)p+4;
                }
            }

//            printf("reloc %s into %x(+%x) @%s\n", name, *p, *p-(int)image, abfd->filename);
        }

        free(loc);

        sect = sect->next;
    }
}

void DtrFile::load_ctor_section(asection* sec) {
    if (!sec) return;
    int size = sec->size / 4;
    int* p = (int*)(origin() + sec->filepos);
    for (int i = 0; i < size; i++) {
        void (*fp)() = (void (*)())*p;
//        dump_func((char*)fp);
        mprotect((void*)(((int)fp+4095) & ~4095 - 4096),
                 4096, PROT_READ | PROT_WRITE | PROT_EXEC);
        fp();
        p++;
    }
}

void DtrFile::load_ctors() {
    if (is_archive()) {
        for_each(arch.begin(), arch.end(), mem_fun(&DtrFile::load_ctors));
        return;
    }

    load_ctor_section(bfd_get_section_by_name(abfd, ".ctors"));

    asection* sec = bfd_get_section_by_name(abfd, ".jcr");
    if (!sec) return;
    int size = sec->size / 4;
    int* p = (int*)(origin() + sec->filepos);
    void (*jrc)(void**) = 0;
    for (vector<void*>::iterator ite = dlhandles.begin();
         ite != dlhandles.end(); ++ite)
    {
        jrc = (void (*)(void**))dlsym(*ite, "_Jv_RegisterClasses");
        if (jrc) break;
    }
    if (!jrc) {
        die("_Jv_RegisterClasses not found");
    }
    void** klasses = (void**)malloc(sizeof(void*)*(size+1));
    for (int i = 0; i < size; i++) {
        void* klass = (void*)*p;
        klasses[i] = klass;
        p++;
    }
    klasses[size] = 0;
    jrc(klasses);
    free(klasses);
}

void DtrFile::add_sym(asymbol* asym) {
    const char* name = bfd_asymbol_name(asym);
    int value = bfd_asymbol_value(asym);
    symbol_info info;

    if (name == NULL || strlen(name) == 0) return;

    bfd_get_symbol_info(abfd, asym, &info);
    if (bfd_is_undefined_symclass(info.type)) return;

    value += origin() + asym->section->filepos;

    fps[name] = value;

    if ((strcmp(name, "test") == 0 ||
         strstr(name, "testEv") == name+strlen(name)-6 ||
         strstr(name, "dtr_test")) &&
        (!strstr(name, "__") || strstr(name, "c_Objc__")))
    {
        tests.push_back(new DtrTest((void (*)())value, this));
    }

//    printf("%08x(+%08x) %s\n", value, value-(int)image, name);
}

void DtrFile::close_bfd() {
    if (abfd) {
        bfd_close(abfd);
        abfd = 0;
    }
}

void DtrFile::find_nearest_line(int p, const char** file, const char** func,
                                unsigned int* line)
{
    if (is_archive()) {
        int min = 0x7fffffff;
        DtrFile* f = 0;
        for (vector<DtrFile*>::iterator ite = arch.begin();
             ite != arch.end(); ++ite)
        {
            int off = p - (*ite)->origin();
            if (off >= 0 && min > off) {
                min = off;
                f = *ite;
            }
        }
        if (!f) {
            die("DtrFile::find_nearest_line???");
        }
        f->find_nearest_line(p, file, func, line);
    }
    else {
        asection* sec = bfd_get_section_by_name(abfd, ".text");
        int off = p - origin() - sec->filepos - 1;
        bfd_find_nearest_line(abfd, sec, syms, off, file, func, line);
    }
}

DtrTest::DtrTest(void (*fp0)(), DtrFile* file0)
    : fp(fp0), file(file0)
{
    mprotect((void*)(((int)fp+4095) & ~4095 - 4096),
             4096, PROT_READ | PROT_WRITE | PROT_EXEC);
}

void DtrTest::run() {
//    dump_func((char*)fp);
    // @@@ signal trap?
    if (!crash) {
        struct sigaction sa;
        memset(&sa, 0, sizeof(sa));
        sa.sa_sigaction = sigtrap;
        sa.sa_flags = SA_RESETHAND | SA_SIGINFO;
        sigaction(SIGHUP, &sa, NULL);
        sigaction(SIGINT, &sa, NULL);
        sigaction(SIGQUIT, &sa, NULL);
        sigaction(SIGILL, &sa, NULL);
        sigaction(SIGABRT, &sa, NULL);
        sigaction(SIGFPE, &sa, NULL);
//        sigaction(SIGKILL, &sa, NULL);
        sigaction(SIGSEGV, &sa, NULL);
        sigaction(SIGPIPE, &sa, NULL);
        sigaction(SIGALRM, &sa, NULL);
        sigaction(SIGTERM, &sa, NULL);
        sigaction(SIGUSR1, &sa, NULL);
        sigaction(SIGUSR2, &sa, NULL);
    }
    try {
        int ret = setjmp(jmpbuf);
        if (ret == 1) {
            failnum++;
        }
        else if (ret == 2) {
            errnum++;
        }
        else {
            fp();
            putchar('.');
        }
    }
    catch (...) {
        errnum++;
        putchar('E');
    }
}

int main(int argc, char* argv[]) {
    failnum = 0;
    errnum = 0;
    assertnum = 0;
#if defined(__linux__)
    dlhandles.push_back(0);
#else
    dlhandles.push_back(RTLD_SELF);
#endif

    /* for glibc...?
    fps["atexit"] = (int)atexit;
    fps["fstat"] = (int)fstat;
    fps["lstat"] = (int)lstat;
    fps["stat"] = (int)stat;
    */

    bfd_init();

    for (int i = 1; i < argc; i++) {
        const char* a = argv[i];
        if (strcmp(a, "-c") == 0) {
            crash = 1;
        }
        else if (strstr(a, "-l") == a) {
            string so(string("lib")+(a+2)+SO_SUFFIX);
            void* h = dlopen(so.c_str(), RTLD_LAZY);
            if (!h) {
                die(so + ": cannot open");
            }
            dlhandles.push_back(h);
        }
        else if (strstr(a, SO_SUFFIX) == a+strlen(a)-strlen(SO_SUFFIX) ||
                 strstr(a, SO_SUFFIX "."))
        {
            void* h = dlopen(a, RTLD_LAZY);
            if (!h) {
                die(string(a) + ": cannot open");
            }
            dlhandles.push_back(h);
        }
        else {
            files.push_back(new DtrFile(a));
        }
    }

    for_each(files.begin(), files.end(), mem_fun(&DtrFile::load_file));
    for_each(files.begin(), files.end(), mem_fun(&DtrFile::load_bfd));
    for_each(files.begin(), files.end(), mem_fun(&DtrFile::load_symtab));
    for_each(files.begin(), files.end(), mem_fun(&DtrFile::load_syms));
    for_each(files.begin(), files.end(), mem_fun(&DtrFile::load_rels));
    for_each(files.begin(), files.end(), mem_fun(&DtrFile::load_ctors));

    clock_t cl = clock();
    printf("Started\n");

    for_each(tests.begin(), tests.end(), mem_fun(&DtrTest::run));

    printf("\nFinished in %f seconds.\n",
           ((float)(clock()-cl)) / CLOCKS_PER_SEC);
    printf("\n%d tests, %d asserts, %d failures, %d errors\n",
           tests.size(), assertnum, failnum, errnum);

    for_each(tests.begin(), tests.end(), delete_fun<DtrTest>);
    for_each(files.begin(), files.end(), delete_fun<DtrFile>);

    return 0;
}
